/* THVGA.H -- Thrillhaus VGA Pack for C/C++ by Matt Gulden
   This file originated at
   http://www.thrillhaus.com/
   It is public domain and may be modified and/or redistributed.
   Includes screen mode calling, pixel plotting, palette changing, point
   returning, wait for retrace */

#include<dos.h>

/* Reduce our workload */
#define word unsigned int
#define byte unsigned char

void videomode(word modenr);
void pixel(word x, word y, byte color);
void palette(byte attribute, byte red, byte green, byte blue);
byte point(word x, word y);
void wait_for_retrace(void);
void blacken(void);
void Set320x133(void);
void Set320x100(void);
void Set320x80(void);

/* This won't work with DJGPP... I'll add that sometime.
   To call mode 13h: videomode(0x13); */
void videomode(word modenr)
{
  asm {
		mov ax, modenr
		int 0x10
		}
}

/* Pretty fast pixel-plotting routine  */
void pixel(word x, word y, byte color)
{
	pokeb(0xa000, (y << 8) + (y << 6) + x, color);
}

void palette(byte attribute, byte red, byte green, byte blue)
{
	outportb(0x3C8, attribute);
	outportb(0x3C9, red);
	outportb(0x3C9, green);
	outportb(0x3C9, blue);
}

/* Returns the color of a pixel */
byte point(word x, word y)
{
	return peekb(0xa000, (y << 8) + (y << 6) + x);
}

/* If you want to reduce flickering in an animation, put this after you
   draw something or before you erase it. */
void wait_for_retrace(void)
{
	while( (inp(0x03DA) & 0x08));
	while(!(inp(0x03DA) & 0x08));
}

void blacken(void)
{
	word COLOR;
	for(COLOR = 0; COLOR < 256; COLOR++) palette(COLOR, 0, 0, 0);
}

/* The next three procedures just modify the screen while in mode 13h */
void set320x133(void)
{
  outportb(0x3D4, 9);
  outportb(0x3D5, 2);
}

void set320x100(void)
{
  outportb(0x3D4, 9);
  outportb(0x3D5, 3);
}

void set320x80(void)
{
  outportb(0x3D4, 9);
  outportb(0x3D5, 4);
}
